<?php

namespace Drupal\webform_qr_code_element\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElementBase;

/**
 * Provides a compound form element for positioned text with formatting.
 *
 * @FormElement("text_position")
 */
class TextPosition extends FormElementBase {

  /**
   * Processes the text position element.
   */
  public static function processTextPosition(&$element, FormStateInterface $form_state, &$complete_form) {
    $value = $element['#default_value'] ?? [];

    $element['#tree'] = TRUE;
    $element['text'] = [
      '#type' => 'textarea',
      '#title' => t('Text'),
      '#required' => TRUE,
      '#default_value' => $value['text'] ?? '',
      '#token_types' => ['webform', 'webform_submission', 'webform_handler', 'site', 'date'],
      '#token_help' => TRUE,
      '#rows' => 2,
    ];

    $element['position'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['text-position-coordinates']],
    ];
    $element['position']['x'] = [
      '#type' => 'number',
      '#title' => t('X position'),
      '#required' => TRUE,
      '#default_value' => $value['x'] ?? 0,
      '#field_suffix' => 'mm',
      '#size' => 8,
      '#step' => 1,
      '#min' => 0,
      '#max' => 210,
    ];
    $element['position']['y'] = [
      '#type' => 'number',
      '#title' => t('Y position'),
      '#required' => TRUE,
      '#default_value' => $value['y'] ?? 0,
      '#field_suffix' => 'mm',
      '#size' => 8,
      '#step' => 1,
      '#min' => 0,
      '#max' => 297,
    ];

    // Font settings container.
    $element['font'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['text-position-font']],
    ];

    $element['font']['font_name'] = [
      '#type' => 'select',
      '#title' => t('Font'),
      '#options' => [
        'Arial' => 'Arial',
        'Courier' => 'Courier',
        'Times' => 'Times',
      ],
      '#default_value' => $value['font_name'] ?? 'Arial',
    ];

    $element['font']['font_size'] = [
      '#type' => 'number',
      '#title' => t('Size'),
      '#required' => TRUE,
      '#default_value' => $value['font_size'] ?? 16,
      '#size' => 6,
      '#step' => 1,
      '#min' => 1,
      '#max' => 72,
    ];

    // Text formatting options.
    $element['formatting'] = [
      '#type' => 'container',
      '#attributes' => ['class' => ['text-position-formatting']],
      '#title' => t('Text Formatting'),
    ];

    $element['formatting']['bold'] = [
      '#type' => 'checkbox',
      '#title' => t('Bold'),
      '#default_value' => $value['bold'] ?? FALSE,
    ];

    $element['formatting']['italic'] = [
      '#type' => 'checkbox',
      '#title' => t('Italic'),
      '#default_value' => $value['italic'] ?? FALSE,
    ];

    $element['formatting']['underlined'] = [
      '#type' => 'checkbox',
      '#title' => t('Underlined'),
      '#default_value' => $value['underlined'] ?? FALSE,
    ];

    $element['#attached']['library'][] = 'webform_qr_code_element/text_position_element';

    return $element;
  }

  /**
   * Reshuffle the submitted data to a flat array for easier upstream use and encapsulation.
   */
  public static function validateTextPosition(&$element, FormStateInterface $form_state) {
    $values = $form_state->getValue($element['#parents']);

    $processed_value = [
      'text' => $values['text'],
      'x' => (int) $values['position']['x'],
      'y' => (int) $values['position']['y'],
      'font_name' => $values['font']['font_name'],
      'font_size' => (int) $values['font']['font_size'],
      'bold' => (bool) $values['formatting']['bold'],
      'italic' => (bool) $values['formatting']['italic'],
      'underlined' => (bool) $values['formatting']['underlined'],
    ];
    foreach (self::getDefaults() as $key => $default) {
      if ($processed_value[$key] === $default) {
        unset($processed_value[$key]);
      }
    }

    $form_state->setValueForElement($element, $processed_value);
  }

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);

    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processTextPosition'],
      ],
      '#element_validate' => [
        [$class, 'validateTextPosition'],
      ],
      '#theme_wrappers' => ['form_element'],
      '#default_value' => $this->getDefaults(),
    ];
  }

  /**
   * Missing values are replaced by their default.
   */
  public static function getSetting(array $element, string $key) {
    return $element[$key] ?? self::getDefaults()[$key];
  }

  /**
   * @return array
   */
  protected static function getDefaults(): array {
    return [
      'text' => '',
      'x' => 0,
      'y' => 0,
      'font_name' => 'Arial',
      'font_size' => 16,
      'bold' => FALSE,
      'italic' => FALSE,
      'underlined' => FALSE,
    ];
  }

}
